# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'nedeľa',
        1: 'pondelok',
        2: 'utorok',
        3: 'streda',
        4: 'štvrtok',
        5: 'piatok',
        6: 'sobota'
    },
    'days_abbrev': {
        0: 'ne',
        1: 'po',
        2: 'ut',
        3: 'st',
        4: 'št',
        5: 'pi',
        6: 'so'
    },

    # Months
    'months': {
        1: 'január',
        2: 'február',
        3: 'marec',
        4: 'apríl',
        5: 'máj',
        6: 'jún',
        7: 'júl',
        8: 'august',
        9: 'september',
        10: 'október',
        11: 'november',
        12: 'december'
    },
    'months_abbrev': {
        1: 'jan',
        2: 'feb',
        3: 'mar',
        4: 'apr',
        5: 'máj',
        6: 'jún',
        7: 'júl',
        8: 'aug',
        9: 'sep',
        10: 'okt',
        11: 'nov',
        12: 'dec'
    },

    # Units of time
    'year': ['rok', '{count} roky', '{count} rokov'],
    'month': ['mesiac', '{count} mesiace', '{count} mesiacov'],
    'week': ['týždeň', '{count} týždne', '{count} týždňov'],
    'day': ['deň', '{count} dni', '{count} dní'],
    'hour': ['hodinu', '{count} hodiny', '{count} hodín'],
    'minute': ['minútu', '{count} minúty', '{count} minút'],
    'second': ['sekundu', '{count} sekundy', '{count} sekúnd'],

    # Relative time
    'ago': 'pred {time}',
    'from_now': 'o {time}',
    'after': '{time} neskôr',
    'before': '{time} predtým',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LT': 'H:mm',
        'LLLL': 'dddd D. MMMM YYYY H:mm',
        'LLL': 'D. MMMM YYYY H:mm',
        'LL': 'D. MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

import sys
import os
import subprocess
import re
from contextlib import contextmanager

from .timezone import Timezone
from .loader import Loader


class LocalTimezone(object):

    _cache = None

    _local_timezone = None

    @classmethod
    def get(cls, force=False):
        if cls._local_timezone is not None:
            return cls._local_timezone

        if cls._cache is None or force:
            name = cls.get_local_tz_name()
            if isinstance(name, Timezone):
                cls._cache = name
            else:
                cls._cache = Timezone.load(cls.get_local_tz_name())

        return cls._cache

    @classmethod
    @contextmanager
    def test(cls, mock):
        """
        Context manager to temporarily set the local_timezone value.

        :type mock: Pendulum or None
        """
        cls.set_local_timezone(mock)

        yield

        cls.set_local_timezone()

    @classmethod
    def set_local_timezone(cls, local_timezone=None):
        cls._local_timezone = local_timezone

    @classmethod
    def get_local_tz_name(cls):
        if sys.platform == 'win32':
            os = 'windows'
        elif 'darwin' in sys.platform:
            os = 'darwin'
        else:
            os = 'unix'

        return getattr(cls, 'get_tz_name_for_{}'.format(os))()

    @classmethod
    def get_tz_name_for_darwin(cls):
        tzname = None
        try:
            output = subprocess.check_output(
                "systemsetup -gettimezone",
                stderr=subprocess.STDOUT
            )

            tzname = output.replace(b'Time Zone: ', b'').strip()
        except OSError:
            pass

        if not tzname:
            # link will be something like /usr/share/zoneinfo/America/Los_Angeles.
            link = os.readlink("/etc/localtime")
            tzname = link[link.rfind("zoneinfo/") + 9:]

        return tzname

    @classmethod
    def get_tz_name_for_windows(cls):
        from tzlocal.win32 import get_localzone_name

        return get_localzone_name()

    @classmethod
    def get_tz_name_for_unix(cls, _root='/'):
        tzenv = os.environ.get('TZ')
        if tzenv:
            try:
                return _tz_from_env(tzenv)
            except ValueError:
                pass

        # Now look for distribution specific configuration files
        # that contain the timezone name.
        tzpath = os.path.join(_root, 'etc/timezone')
        if os.path.exists(tzpath):
            with open(tzpath, 'rb') as tzfile:
                data = tzfile.read()

                # Issue #3 was that /etc/timezone was a zoneinfo file.
                # That's a misconfiguration, but we need to handle it gracefully:
                if data[:5] != 'TZif2':
                    etctz = data.strip().decode()
                    # Get rid of host definitions and comments:
                    if ' ' in etctz:
                        etctz, dummy = etctz.split(' ', 1)
                    if '#' in etctz:
                        etctz, dummy = etctz.split('#', 1)

                    return etctz.replace(' ', '_')

        # CentOS has a ZONE setting in /etc/sysconfig/clock,
        # OpenSUSE has a TIMEZONE setting in /etc/sysconfig/clock and
        # Gentoo has a TIMEZONE setting in /etc/conf.d/clock
        # We look through these files for a timezone:

        zone_re = re.compile('\s*ZONE\s*=\s*\"')
        timezone_re = re.compile('\s*TIMEZONE\s*=\s*\"')
        end_re = re.compile('\"')

        for filename in ('etc/sysconfig/clock', 'etc/conf.d/clock'):
            tzpath = os.path.join(_root, filename)
            if not os.path.exists(tzpath):
                continue
            with open(tzpath, 'rt') as tzfile:
                data = tzfile.readlines()

            for line in data:
                # Look for the ZONE= setting.
                match = zone_re.match(line)
                if match is None:
                    # No ZONE= setting. Look for the TIMEZONE= setting.
                    match = timezone_re.match(line)
                if match is not None:
                    # Some setting existed
                    line = line[match.end():]
                    etctz = line[:end_re.search(line).start()]

                    # We found a timezone
                    return etctz.replace(' ', '_')

        # systemd distributions use symlinks that include the zone name,
        # see manpage of localtime(5) and timedatectl(1)
        tzpath = os.path.join(_root, 'etc/localtime')
        if os.path.exists(tzpath) and os.path.islink(tzpath):
            tzpath = os.path.realpath(tzpath)
            parts = tzpath.split('/')[-2:]
            while parts:
                tzpath = '/'.join(parts)
                try:
                    return Timezone.load(tzpath)
                except (ValueError, IOError, OSError):
                    pass

                parts.pop(0)

        # No explicit setting existed. Use localtime
        for filename in ('etc/localtime', 'usr/local/etc/localtime'):
            tzpath = os.path.join(_root, filename)

            if not os.path.exists(tzpath):
                continue
            return Timezone('', *Loader.load_from_file(tzpath))

        raise RuntimeError('Can not find any timezone configuration')


def _tz_from_env(tzenv):
    if tzenv[0] == ':':
        tzenv = tzenv[1:]

    # TZ specifies a file
    if os.path.exists(tzenv):
        return Timezone('', *Loader.load(tzenv))

    # TZ specifies a zoneinfo zone.
    try:
        tz = Timezone.load(tzenv)

        return tz
    except ValueError:
        raise



# -*- coding: utf-8 -*-


from datetime import datetime, timedelta
from struct import unpack, calcsize
from pytzdata import tz_file
from pytzdata.exceptions import TimezoneNotFound

from .. import _compat
from ..helpers import local_time
from .transition import Transition
from .transition_type import TransitionType


def _byte_string(s):
    """Cast a string or byte string to an ASCII byte string."""
    return s.encode('US-ASCII')

_NULL = _byte_string('\0')


def _std_string(s):
    """Cast a string or byte string to an ASCII string."""
    return str(s.decode('US-ASCII'))


class Loader(object):

    @classmethod
    def load(cls, name):
        name = _compat.decode(name)
        try:
            with tz_file(name) as f:
                return cls._load(f)
        except TimezoneNotFound:
            raise ValueError('Unknown timezone [{}]'.format(name))

    @classmethod
    def load_from_file(cls, filepath):
        try:
            with open(filepath, 'rb') as f:
                return cls._load(f)
        except _compat.FileNotFoundError:
            raise ValueError('Unable to load file [{}]'.format(filepath))

    @classmethod
    def _load(cls, fp):
        head_fmt = '>4s c 15x 6l'
        head_size = calcsize(head_fmt)
        (magic, fmt, ttisgmtcnt, ttisstdcnt, leapcnt, timecnt,
         typecnt, charcnt) = unpack(head_fmt, fp.read(head_size))

        # Make sure it is a tzfile(5) file
        assert magic == _byte_string('TZif'), 'Got magic %s' % repr(magic)

        # Read out the transition times,
        # localtime indices and ttinfo structures.
        data_fmt = '>%(timecnt)dl %(timecnt)dB %(ttinfo)s %(charcnt)ds' % dict(
            timecnt=timecnt, ttinfo='lBB' * typecnt, charcnt=charcnt)
        data_size = calcsize(data_fmt)
        data = unpack(data_fmt, fp.read(data_size))

        # make sure we unpacked the right number of values
        assert len(data) == 2 * timecnt + 3 * typecnt + 1
        transition_times = tuple(trans for trans in data[:timecnt])
        lindexes = tuple(data[timecnt:2 * timecnt])
        ttinfo_raw = data[2 * timecnt:-1]
        tznames_raw = data[-1]
        del data

        # Process ttinfo into separate structs
        transition_types = tuple()
        tznames = {}
        i = 0
        while i < len(ttinfo_raw):
            # have we looked up this timezone name yet?
            tzname_offset = ttinfo_raw[i + 2]
            if tzname_offset not in tznames:
                nul = tznames_raw.find(_NULL, tzname_offset)
                if nul < 0:
                    nul = len(tznames_raw)
                tznames[tzname_offset] = _std_string(
                    tznames_raw[tzname_offset:nul])
            transition_types += (
                TransitionType(
                    ttinfo_raw[i], bool(ttinfo_raw[i + 1]),
                    tznames[tzname_offset]
                ),
            )
            i += 3

        # Now build the timezone object
        transitions = tuple()
        tzinfos = tuple()

        if not transition_times:
            if transition_types:
                transitions += (Transition(0, 0, datetime(1970, 1, 1), datetime(1970, 1, 1), 0),)
                tzinfos += ((
                    transition_types[0].utc_offset,
                    transition_types[0].is_dst,
                    None,
                    transition_types[0].abbrev
                ),)
        else:
            # calculate transition info
            tr = None
            for i in range(len(transition_times)):
                # We retrieve transition types indexes
                transition_type_index = lindexes[i]

                if i == 0:
                    pre_transition_type_index = lindexes[i]
                else:
                    pre_transition_type_index = lindexes[i - 1]

                pre_transition_type = transition_types[pre_transition_type_index]
                transition_type = transition_types[transition_type_index]

                # We calculate local times based on the transition types
                # we retrieved
                pre_time = datetime(
                    *local_time(transition_times[i],
                                pre_transition_type.utc_offset,
                                0)
                )
                time = datetime(
                    *local_time(transition_times[i],
                                transition_type.utc_offset,
                                0)
                )

                # We build the tzinfo information as tuples
                # and retrieve their index to store them
                # in the transition.
                # If they do not already exist we add them
                # to the tzinfos tuple to look them up later
                # if necessary.
                dst = tr.time - tr.pre_time if tr else None
                pre_tzinfo = (
                    pre_transition_type.utc_offset,
                    pre_transition_type.is_dst,
                    dst if dst != timedelta() else None,
                    pre_transition_type.abbrev
                )

                try:
                    pre_tzinfo_index = tzinfos.index(pre_tzinfo)
                except ValueError:
                    tzinfos += (pre_tzinfo,)

                    pre_tzinfo_index = len(tzinfos) - 1

                dst = time - pre_time
                tzinfo = (
                    transition_type.utc_offset,
                    transition_type.is_dst,
                    dst if dst != timedelta() else None,
                    transition_type.abbrev
                )

                try:
                    tzinfo_index = tzinfos.index(tzinfo)
                except ValueError:
                    tzinfos += (tzinfo,)

                    tzinfo_index = len(tzinfos) - 1

                # Finally, we build the transition instance.
                tr = Transition(
                    transition_times[i],
                    tzinfo_index,
                    pre_time,
                    time,
                    pre_tzinfo_index
                )

                transitions += (tr,)

        # Determine the before-first-transition type
        default_tzinfo_index = 0
        if transitions:
            index = 0
            if tzinfos[0][2]:
                index = transitions[0].tzinfo_index
                while index != 0 and tzinfos[index][2]:
                    index -= 1

            while index != len(tzinfos) and tzinfos[index][2]:
                index += 1

            if index != len(transitions):
                default_tzinfo_index = index

        return (
            transitions,
            tzinfos,
            default_tzinfo_index,
            tuple(map(lambda tr: tr.utc_time, transitions))
        )



# -*- coding: utf-8 -*-


def meridian(hour, minute):
    hm = hour * 100 + minute

    if hm < 600:
        return '凌晨'
    elif hm < 900:
        return '早上'
    elif hm < 1130:
        return '上午'
    elif hm < 1230:
        return '中午'
    elif hm < 1800:
        return '下午'

    return '晚上'


translations = {
    # Days
    'days': {
        0: '星期日',
        1: '星期一',
        2: '星期二',
        3: '星期三',
        4: '星期四',
        5: '星期五',
        6: '星期六'
    },
    'days_abbrev': {
        0: '日',
        1: '一',
        2: '二',
        3: '三',
        4: '四',
        5: '五',
        6: '六'
    },

    # Months
    'months': {
        1: '一月',
        2: '二月',
        3: '三月',
        4: '四月',
        5: '五月',
        6: '六月',
        7: '七月',
        8: '八月',
        9: '九月',
        10: '十月',
        11: '十一月',
        12: '十二月',
    },
    'months_abbrev': {
        1: '1月',
        2: '2月',
        3: '3月',
        4: '4月',
        5: '5月',
        6: '6月',
        7: '7月',
        8: '8月',
        9: '9月',
        10: '10月',
        11: '11月',
        12: '12月',
    },

    # Units of time
    'year': '{count}年',
    'month': '{count}个月',
    'week': '{count}周',
    'day': '{count}天',
    'hour': '{count}小时',
    'minute': '{count}分钟',
    'second': '{count}秒',

    # Relative time
    'ago': '{time}前',
    'from_now': '距现在{time}',
    'after': '{time}后',
    'before': '{time}前',

    # Meridians
    'meridian': lambda time: meridian(*time),

    # Date formats
    'date_formats': {
        'LTS': 'Ah点m分s秒',
        'LT': 'Ah点mm分',
        'LLLL': 'YYYY年MMMD日ddddAh点mm分',
        'LLL': 'YYYY年MMMD日Ah点mm分',
        'LL': 'YYYY年MMMD日',
        'L': 'YYYY-MM-DD',
    },
}




# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Domingo',
        1: 'Segunda-feira',
        2: 'Terça-feira',
        3: 'Quarta-feira',
        4: 'Quinta-feira',
        5: 'Sexta-feira',
        6: 'Sábado'
    },
    'days_abbrev': {
        0: 'Dom',
        1: 'Seg',
        2: 'Ter',
        3: 'Qua',
        4: 'Qui',
        5: 'Sex',
        6: 'Sab'
    },

    # Months
    'months': {
        1: 'Janeiro',
        2: 'Fevereiro',
        3: 'Março',
        4: 'Abril',
        5: 'Maio',
        6: 'Junho',
        7: 'Julho',
        8: 'Agosto',
        9: 'Setembro',
        10: 'Outubro',
        11: 'Novembro',
        12: 'Dezembro',
    },
    'months_abbrev': {
        1: 'Jan',
        2: 'Fev',
        3: 'Mar',
        4: 'Abr',
        5: 'Maio',
        6: 'Jun',
        7: 'Jul',
        8: 'Ago',
        9: 'Set',
        10: 'Out',
        11: 'Nov',
        12: 'Dez',
    },

    # Units of time
    'year': ['{count} ano', '{count} anos'],
    'month': ['{count} mês', '{count} meses'],
    'week': ['{count} semana', '{count} semanas'],
    'day': ['{count} dia', '{count} dias'],
    'hour': ['{count} hora', '{count} horas'],
    'minute': ['{count} minuto', '{count} minutos'],
    'second': ['{count} segundo', '{count} segundos'],

    # Relative time
    'ago': 'há {time}',
    'from_now': 'dentro de {time}',
    'after': 'após {time}',
    'before': '{time} atrás',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D [de] MMMM [de] YYYY [às] HH:mm',
        'LLL': 'D [de] MMMM [de] YYYY [às] HH:mm',
        'LL': 'D [de] MMMM [de] YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

from .mixins.default import (
    TestableMixin, FormattableMixing, TranslatableMixin
)
from .pendulum import Pendulum
from .date import Date
from .time import Time


class Global(TestableMixin, FormattableMixing, TranslatableMixin):

    @classmethod
    def set_test_now(cls, test_now=None):
        cls._test_now = test_now

        Pendulum.set_test_now(test_now)
        Date.set_test_now(test_now)
        Time.set_test_now(test_now)

    @classmethod
    def set_formatter(cls, formatter=None):
        super(Global, cls).set_formatter(formatter)

        Pendulum.set_formatter(formatter)
        Date.set_formatter(formatter)
        Time.set_formatter(formatter)

    @classmethod
    def set_locale(cls, locale):
        super(Global, cls).set_locale(locale)

        Pendulum.set_locale(locale)
        Date.set_locale(locale)
        Time.set_locale(locale)



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'الاثنين',
        1: 'الأحد',
        2: 'السبت',
        3: 'الجمعة',
        4: 'الخميس',
        5: 'الأربعاء',
        6: 'الثلاثاء'
    },
    'days_abbrev': {
        0: 'اثنين',
        1: 'أحد',
        2: 'سبت',
        3: 'جمعة',
        4: 'خميس',
        5: 'أربعاء',
        6: 'ثلاثاء'
    },

    # Months
    'months': {
        1: 'يوليو',
        2: 'يونيو',
        3: 'مايو',
        4: 'أبريل',
        5: 'مارس',
        6: 'فبراير',
        7: 'يناير',
        8: 'ديسمبر',
        9: 'نوفمبر',
        10: 'أكتوبر',
        11: 'سبتمبر',
        12: 'أغسطس',
    },
    'months_abbrev': {
        1: 'يوليو',
        2: 'يونيو',
        3: 'مايو',
        4: 'أبريل',
        5: 'مارس',
        6: 'فبراير',
        7: 'يناير',
        8: 'ديسمبر',
        9: 'نوفمبر',
        10: 'أكتوبر',
        11: 'سبتمبر',
        12: 'أغسطس',
    },

    # Units of time
    'year': {
        0: 'سنة',
        1: 'سنة',
        2: 'سنتين',
        (3, 10): '{count} سنوات',
        'default': '{count} سنة'
    },
    'month': {
        0: 'شهر',
        1: ' شهر',
        2: 'شهرين',
        (3, 10): '{count} أشهر',
        'default': '{count} شهر'
    },
    'week': {
        0: 'إسبوع',
        1: 'إسبوع',
        2: 'إسبوعين',
        (3, 10): '{count} أسابيع',
        'default': '{count} إسبوع'
    },
    'day': {
        0: 'يوم',
        1: 'يوم',
        2: 'يومين',
        (3, 10): '{count} أيام',
        'default': ' يوم'
    },
    'hour': {
        0: 'ساعة',
        1: 'ساعة',
        2: 'ساعتين',
        (3, 10): '{count} ساعات',
        'default': '{count} ساعة'
    },
    'minute': {
        0: 'دقيقة',
        1: 'دقيقة',
        2: 'دقيقتين',
        (3, 10): '{count} دقائق',
        'default': '{count} دقيقة'
    },
    'second': {
        0: 'ثانية',
        1: 'ثانية',
        2: 'ثانيتين',
        (3, 10): '{count} ثوان',
        'default': '{count} ثانية'
    },

    # Relative time
    'ago': 'منذ {time}',
    'from_now': 'من الآن {time}',
    'after': 'بعد {time}',
    'before': 'قبل {time}',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'D/\u200FM/\u200FYYYY',
    },
}



# -*- coding: utf-8 -*-

from .formatter import Formatter
from .classic_formatter import ClassicFormatter
from .alternative_formatter import AlternativeFormatter


FORMATTERS = {
    'classic': ClassicFormatter(),
    'alternative': AlternativeFormatter(),
}


def register_formatter(name, formatter):
    """
    Register a new formatter.

    :param name: The name of the formatter.
    :type name: str

    :param formatter: The formatter instance
    :type formatter: Formatter

    :rtype: None
    """
    if name in FORMATTERS:
        raise ValueError('Formatter [{}] already exists'.format(name))

    if not isinstance(formatter, Formatter):
        raise ValueError(
            'The formatter instance '
            'must be an instance of Formatter'
        )

    FORMATTERS[name] = formatter



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'неділя',
        1: 'понеділок',
        2: 'вівторок',
        3: 'середа',
        4: 'четвер',
        5: 'п\'ятниця',
        6: 'субота'
    },
    'days_abbrev': {
        0: 'Нд',
        1: 'Пн',
        2: 'Вт',
        3: 'Ср',
        4: 'Чт',
        5: 'Пт',
        6: 'Сб'
    },

    # Months
    'months': {
        1: 'січня',
        2: 'лютого',
        3: 'березня',
        4: 'квітня',
        5: 'травня',
        6: 'червня',
        7: 'липня',
        8: 'серпня',
        9: 'вересня',
        10: 'жовтня',
        11: 'листопада',
        12: 'грудня'
    },
    'months_abbrev': {
        1: 'січ',
        2: 'лют',
        3: 'бер',
        4: 'квіт',
        5: 'трав',
        6: 'черв',
        7: 'лип',
        8: 'серп',
        9: 'вер',
        10: 'жовт',
        11: 'лист',
        12: 'груд'
    },

    # Units of time
    'year': ['{count} рік', '{count} роки', '{count} років'],
    'month': ['{count} місяць', '{count} місяці', '{count} місяців'],
    'week': ['{count} тиждень', '{count} тижні', '{count} тижнів'],
    'day': ['{count} день', '{count} дні', '{count} днів'],
    'hour': ['{count} година', '{count} години', '{count} годин'],
    'minute': ['{count} хвилину', '{count} хвилини', '{count} хвилин'],
    'second': ['{count} секунду', '{count} секунди', '{count} секунд'],

    # Relative time
    'ago': '{time} назад',
    'from_now': 'через {time}',
    'after': '{time} після',
    'before': '{time} до',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY р., HH:mm',
        'LLL': 'D MMMM YYYY р., HH:mm',
        'LL': 'D MMMM YYYY р.',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'søndag',
        1: 'mandag',
        2: 'tirsdag',
        3: 'onsdag',
        4: 'torsdag',
        5: 'fredag',
        6: 'lørdag'
    },
    'days_abbrev': {
        0: 'sø.',
        1: 'ma.',
        2: 'ti.',
        3: 'on.',
        4: 'to.',
        5: 'fr.',
        6: 'lø.'
    },

    # Months
    'months': {
        1: 'januar',
        2: 'februar',
        3: 'mars',
        4: 'april',
        5: 'mai',
        6: 'juni',
        7: 'juli',
        8: 'august',
        9: 'september',
        10: 'oktober',
        11: 'november',
        12: 'desember'
    },
    'months_abbrev': {
        1: 'jan.',
        2: 'feb.',
        3: 'mars',
        4: 'april',
        5: 'mai',
        6: 'juni',
        7: 'juli',
        8: 'aug.',
        9: 'sep.',
        10: 'okt.',
        11: 'nov.',
        12: 'des.'
    },

    # Units of time
    'year': ['{count} år', '{count} år'],
    'month': ['{count} måned', '{count} måneder'],
    'week': ['{count} uke', '{count} uker'],
    'day': ['{count} dag', '{count} dager'],
    'hour': ['{count} time', '{count} timer'],
    'minute': ['{count} minutt', '{count} minutter'],
    'second': ['{count} sekund', '{count} sekunder'],

    # Relative time
    'ago': '{time} siden',
    'from_now': 'om {time}',
    'after': '{time} etter',
    'before': '{time} før',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd D. MMMM YYYY [kl.] HH:mm',
        'LLL': 'D. MMMM YYYY [kl.] H:mm',
        'LL': 'D. MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

from datetime import datetime, tzinfo
from bisect import bisect_right

from ..constants import SECONDS_PER_DAY
from .loader import Loader
from .timezone_info import TimezoneInfo, UTC
from ..helpers import local_time as _local_time
from .transition_type import TransitionType
from .exceptions import NonExistingTime, AmbiguousTime


class Timezone(tzinfo):
    """
    Represents a named timezone.

    It inherits from tzinfo in order to be passed to astimezone().
    """

    _cache = {}

    PRE_TRANSITION = 'pre'
    POST_TRANSITION = 'post'
    TRANSITION_ERROR = 'error'

    def __init__(self, name, transitions,
                 tzinfos,
                 default_tzinfo_index,
                 utc_transition_times):
        """
        Constructor.

        :param name: The name of the timezone.
        :type name: str

        :param transitions: The timezone transitions
        :type transitions: tuple

        :param tzinfos: The timezone information.
        :type tzinfos: tuple

        :param default_tzinfo_index: The default TimezoneInfo index.
        :type default_tzinfo_index: int

        :param utc_transition_times: Timestamps of transition times (UTC)
        :type utc_transition_times: list
        """
        self._name = name
        self._transitions = transitions
        self.__tzinfos = tzinfos
        self._tzinfos = tuple(
            map(lambda tzinfo: TimezoneInfo(self, *tzinfo), tzinfos)
        )
        self._default_tzinfo_index = default_tzinfo_index
        self._utc_transition_times = utc_transition_times
        self._local_hint = {}

    @property
    def name(self):
        return self._name

    @property
    def transitions(self):
        return self._transitions

    @property
    def tzinfos(self):
        return self._tzinfos

    @classmethod
    def load(cls, name):
        """
        Loads a timezone with the given name or
        returns it from the cache.

        :param name: The name of the timezone
        :type name: str or int

        :rtype: Timezone
        """
        # Shortcut to UTC
        if name.upper() == 'UTC':
            return UTCTimezone

        if name not in cls._cache:
            (transitions,
             tzinfos,
             default_tzinfo_index,
             utc_transition_times) = Loader.load(name)

            zone = cls(name,
                       transitions,
                       tzinfos,
                       default_tzinfo_index,
                       utc_transition_times)

            cls._cache[name] = zone

        return cls._cache[name]

    def convert(self, dt, dst_rule=None):
        """
        Converts or normalizes a datetime.

        If there is no tzinfo set on the datetime, local time will be assumed
        and normalization will occur.

        Otherwise, it will convert the datetime to local time.
        """
        if dt.tzinfo is None:
            # we assume local time
            converted = self._normalize(dt, dst_rule=dst_rule)
        else:
            converted = self._convert(dt)

        if not isinstance(converted, tuple):
            return converted

        return dt.__class__(*converted[0], **converted[1])

    def datetime(self, year, month, day,
                 hour=0, minute=0, second=0, microsecond=0):
        """
        Creates a new datetime object for the current timezone.

        :param year: The year
        :type year: int

        :param month: The month
        :type month: int

        :param day: The day
        :type day: int

        :param hour: The hour
        :type hour: int

        :param minute: The minute
        :type minute: int

        :param second: The second
        :type second: int

        :param microsecond: The microsecond
        :type microsecond: int

        :rtype: datetime
        """
        dt = datetime(year, month, day, hour, minute, second, microsecond)

        return self.convert(dt, dst_rule=self.POST_TRANSITION)

    def _normalize(self, dt, dst_rule=None):
        # if tzinfo is set, something wrong happened
        if dt.tzinfo is not None:
            raise ValueError(
                'A datetime with a tzinfo cannot be normalized. '
                'Use _convert() instead.'
            )

        # fold attribute (Python 3.6)?
        # We use it to determine the DST rule if none has been specified.
        fold = None
        if dst_rule is None:
            if hasattr(dt, 'fold'):
                fold = dt.fold
                if dt.fold == 1:
                    dst_rule = self.POST_TRANSITION
                else:
                    dst_rule = self.PRE_TRANSITION
            else:
                dst_rule = self.POST_TRANSITION

        if not self._transitions:
            # Use the default offset
            offset = self._tzinfos[self._default_tzinfo_index].offset
            unix_time = (dt - datetime(1970, 1, 1)).total_seconds() - offset

            return self._to_local_time(
                unix_time, dt.microsecond, self._default_tzinfo_index,
                fold
            )

        # Find the first transition after our target date/time
        begin = self._transitions[0]
        end = self._transitions[-1]

        if dt < begin.time:
            tr = begin
        elif dt >= end.time:
            tr = end
        else:
            idx = self._find_transition_index(dt)
            tr = self._transitions[idx]

            if idx > 0:
                pre_tr = self._transitions[idx - 1]

                # DST -> No DST
                if dt <= pre_tr.pre_time:
                    tr = pre_tr

        tzinfo_index = tr._tzinfo_index
        if tr is begin:
            if tr.pre_time >= dt:
                # Before first transition, so use the default offset.
                offset = self._tzinfos[self._default_tzinfo_index].offset
                unix_time = (dt - datetime(1970, 1, 1)).total_seconds() - offset

                return self._to_local_time(
                    unix_time, dt.microsecond, self._default_tzinfo_index,
                    fold
                )
            else:
                if begin is end:
                    # We only have one transition
                    offset = self._tzinfos[tzinfo_index].offset
                    unix_time = (dt - datetime(1970, 1, 1)).total_seconds() - offset
                else:
                    unix_time, tzinfo_index = self._get_time(dt, tr, dst_rule)
        elif tr is end:
            if tr.pre_time < dt:
                # After the last transition.
                unix_time = tr.unix_time + (dt - tr.time).total_seconds()
            else:
                unix_time, tzinfo_index = self._get_time(dt, tr, dst_rule)
        else:
            unix_time, tzinfo_index = self._get_time(dt, tr, dst_rule)

        return self._to_local_time(
            unix_time, dt.microsecond, tzinfo_index,
            fold
        )

    def _get_time(self, dt, tr, dst_rule):
        tzinfo_index = tr._tzinfo_index

        if tr.pre_time <= dt < tr.time:
            # tr.pre_time <= dt < tr.time
            # Skipped time
            if dst_rule == self.TRANSITION_ERROR:
                raise NonExistingTime(dt)
            elif dst_rule == self.PRE_TRANSITION:
                # We do not apply the transition
                (unix_time,
                 tzinfo_index) = self._get_previous_transition_time(tr, dt,
                                                                    skipped=True)
            else:
                unix_time = tr.unix_time - (tr.pre_time - dt).total_seconds()
        elif tr.time <= dt <= tr.pre_time:
            # tr.time <= dt <= tr.pre_time
            # Repeated time
            if dst_rule == self.TRANSITION_ERROR:
                raise AmbiguousTime(dt)
            elif dst_rule == self.PRE_TRANSITION:
                # We do not apply the transition
                (unix_time,
                 tzinfo_index) = self._get_previous_transition_time(tr, dt)
            else:
                unix_time = tr.unix_time + (dt - tr.time).total_seconds()
        else:
            # In between transitions
            # The actual transition type is the previous transition one
            (unix_time,
             tzinfo_index) = self._get_previous_transition_time(tr, dt)

        return unix_time, tzinfo_index

    def _convert(self, dt):
        """
        Converts a timezone-aware datetime to local time.

        :param dt: The datetime to convert.
        :type dt: datetime
        """
        # if tzinfo is not set, something wrong happened
        if dt.tzinfo is None:
            raise ValueError(
                'A datetime without a tzinfo cannot be converted. '
                'Use _normalize() instead.'
            )

        return dt.astimezone(self)

    def _to_local_time(self, unix_time, microseconds, tzinfo_index, fold):
        """
        Returns the local time information
        as a tuple of date, time and keyword arguments (tzinfo and fold),
        given a unix time and a tzinfo index.

        :param unix_time: The timestamp of the transition time (UTC)
        :type unix_time: int

        :param microseconds: The microseconds value
        :type microseconds: int

        :param tzinfo_index: The index of the TimezoneInfo instance
        :type tzinfo_index: int

        :param fold: The fold value (if None, will be discarded)
        :type fold: int or None
        """
        tzinfo = self._tzinfos[tzinfo_index]

        local_time = _local_time(
            unix_time,
            tzinfo.offset,
            microseconds
        )

        keywords = {
            'tzinfo': tzinfo
        }

        if fold is not None:
            keywords['fold'] = fold

        return local_time, keywords

    def _get_diff(self, dt1, dt2):
        diff = dt2 - dt1

        return diff.days * SECONDS_PER_DAY + diff.seconds

    def _find_transition_index(self, dt, prop='_time'):
        lo, hi = 0, len(self._transitions)
        hint = self._local_hint.get(prop)
        if hint:
            if dt == hint[0]:
                return hint[1]
            elif dt < hint[0]:
                hi = hint[1]
            else:
                lo = hint[1]

        while lo < hi:
            mid = (lo + hi) // 2
            if dt < getattr(self._transitions[mid], prop):
                hi = mid
            else:
                lo = mid + 1

        self._local_hint[prop] = (dt, lo)

        return lo

    def _get_previous_transition_time(self, tr, dt, skipped=False):
        """
        Returns the time before the transition
        as a (unix_time, tzinfo_index) tuple.

        :param tr: The transition
        :type tr: Transition

        :param dt: The datetime
        :type dt: datetime

        :param skipped: Whether we are in a gap or not
        :type skipped: bool

        :rtype: tuple
        """
        diff = self._get_diff(tr.pre_time, dt)
        if -1 < diff < 0 and tr.unix_time < 0:
            diff -= 1

        tzinfo_index = tr.pre_tzinfo_index

        unix_time = tr.unix_time + diff

        if skipped:
            # If skipped time, we round down
            # Only occurs when PRE_TRANSITION is used
            forward = (self._tzinfos[tr.tzinfo_index].offset
                       - self._tzinfos[tzinfo_index].offset)
            unix_time -= forward

        return unix_time, tzinfo_index

    def tzname(self, dt):
        if dt is None:
            return None

        if dt.tzinfo is self:
            dt = self.convert(dt.replace(tzinfo=None))
        else:
            dt = self.convert(dt)

        return dt.tzinfo.abbrev

    def utcoffset(self, dt):
        if dt is None:
            return None

        if dt.tzinfo is self:
            dt = self.convert(dt.replace(tzinfo=None))
        else:
            dt = self.convert(dt)

        return dt.tzinfo.adjusted_offset

    def dst(self, dt):
        if dt is None:
            return None

        if dt.tzinfo is self:
            dt = self.convert(dt.replace(tzinfo=None))
        else:
            dt = self.convert(dt)

        return dt.tzinfo.dst_

    def fromutc(self, dt):
        dt = dt.replace(tzinfo=None)

        idx = self._find_utc_index(dt)
        tr = self._transitions[idx]
        tzinfo = self._tzinfos[tr._tzinfo_index]

        return (dt + tzinfo.adjusted_offset).replace(tzinfo=tzinfo)

    def _find_utc_index(self, dt):
        lo, hi = 0, len(self._utc_transition_times)
        hint = self._local_hint.get('_utc')
        if hint:
            if dt == hint[0]:
                return hint[1]
            elif dt < hint[0]:
                hi = hint[1] + 1
            else:
                lo = hint[1]

        idx = max(0, bisect_right(self._utc_transition_times, dt, lo, hi) - 1)

        self._local_hint['_utc'] = (dt, idx)

        return idx

    def __repr__(self):
        return '<Timezone [{}]>'.format(self._name)

    def __getinitargs__(self):
        return (self._name, self._transitions,
                self.__tzinfos, self._default_tzinfo_index,
                self._utc_transition_times)


class FixedTimezone(Timezone):
    """
    A timezone that has a fixed offset to UTC.
    """

    _cache = {}

    def __init__(self, offset, name=None, transition_type=None):
        """
        :param offset: offset to UTC in seconds.
        :type offset: int
        """
        sign = '-' if offset < 0 else '+'

        minutes = offset / 60
        hour, minute = divmod(abs(int(minutes)), 60)

        if not name:
            name = '{0}{1:02d}:{2:02d}'.format(sign, hour, minute)

        if not transition_type:
            transition_type = TransitionType(int(offset), False, '')

        super(FixedTimezone, self).__init__(name, [], [], 0, (datetime(1970, 1, 1),))

        self._tzinfos = (
            TimezoneInfo(
                self,
                transition_type.utc_offset,
                transition_type.is_dst,
                None,
                transition_type.abbrev,
            ),
        )
        self._tzinfo = self._tzinfos[0]

        self._offset = offset

    @classmethod
    def load(cls, name):
        if name not in cls._cache:
            cls._cache[name] = cls(name)

        return cls._cache[name]

    def _normalize(self, dt, dst_rule=Timezone.POST_TRANSITION):
        return dt.replace(tzinfo=self._tzinfo)

    def utcoffset(self, dt):
        if dt is None:
            return None

        return self._tzinfo.adjusted_offset

    def dst(self, dt):
        if dt is None:
            return None

        return self._tzinfo.dst(dt)

    def fromutc(self, dt):
        dt = dt.replace(tzinfo=None)

        return (dt + self._tzinfo.adjusted_offset).replace(tzinfo=self._tzinfo)

    def __getinitargs__(self):
        return (self._offset, )


class _UTC(FixedTimezone):

    def __init__(self):
        super(_UTC, self).__init__(0, 'UTC', TransitionType(0, False, 'UTC'))

        UTC._tz = self

    def fromutc(self, dt):
        return dt.replace(tzinfo=UTC)

    def __getinitargs__(self):
        return ()


UTCTimezone = _UTC()



# -*- coding: utf-8 -*-

import re
import copy
import warnings

from datetime import datetime, date, time
from dateutil import parser

from ..helpers import parse_iso8601, week_day, days_in_year
from .exceptions import ParserError


class Parser(object):
    """
    Parser which parses common formats (like RFC3339 and ISO8601).
    """

    COMMON = re.compile(
        # Date (optional)
        '^'
        '(?P<date>'
        '    (?P<classic>'  # Classic date (YYYY-MM-DD) or ordinal (YYYY-DDD)
        '        (?P<year>\d{4})'  # Year
        '        (?P<monthday>'
        '            (?P<monthsep>[-/:])?(?P<month>\d{2})'  # Month (optional)
        '            ((?P<daysep>[-/:])?(?P<day>\d{1,2}))?'  # Day (optional)
        '        )?'
        '    )'
        '    |'
        '    (?P<isocalendar>'  # Calendar date (2016-W05 or 2016-W05-5)
        '        (?P<isoyear>\d{4})'  # Year
        '        -?'  # Separator (optional)
        '        W'  # W separator
        '        (?P<isoweek>\d{2})'  # Week number
        '        -?'  # Separator (optional)
        '        (?P<isoweekday>\d)?'  # Weekday (optional)
        '    )'
        ')?'

        # Time (optional)
        '(?P<time>'
        '    (?P<timesep>T|\ )?'  # Separator (T or space)
        '    (?P<hour>\d{1,2}):?(?P<minute>\d{1,2})?:?(?P<second>\d{1,2})?'  # HH:mm:ss (optional mm and ss)
        # Subsecond part (optional)
        '    (?P<subsecondsection>'
        '        (?:\.|,)'  # Subsecond separator (optional)
        '        (?P<subsecond>\d{1,9})'  # Subsecond
        '    )?'
        # Timezone offset
        '    (?P<tz>'
        '        (?:-|\+)\d{2}:?(?:\d{2})?|Z'  # Offset (+HH:mm or +HHmm or +HH or Z)
        '    )?'
        ')?'
        '$',
        re.VERBOSE
    )

    DEFAULT_OPTIONS = {
        'day_first': False,
        'year_first': True,
        'exact': False,
        'now': None
    }

    def __init__(self, **options):
        if 'strict' in options:
            warnings.warn(
                'The "strict" keyword when parsing will have '
                'another meaning in version 2.0. Use "exact" instead.',
                DeprecationWarning,
                stacklevel=2
            )

            options['exact'] = options['strict']

        self._options = copy.copy(self.DEFAULT_OPTIONS)
        self._options.update(options)

    def is_strict(self):
        warnings.warn(
            'is_strict() is deprecated. Use is_exact() instead.',
            DeprecationWarning,
            stacklevel=2
        )

        return self.is_exact()

    def is_exact(self):
        return self._options['exact']

    def now(self):
        return self._options['now'] or datetime.now()

    def parse_common(self, text):
        """
        Tries to parse the string as a common datetime format.

        :param text: The string to parse.
        :type text: str

        :rtype: dict or None
        """
        m = self.COMMON.match(text)
        parsed = {}
        ambiguous_date = False
        has_date = False

        if m:
            if m.group('date'):
                # A date has been specified
                has_date = True

                if m.group('isocalendar'):
                    # We have a ISO 8601 string defined
                    # by week number
                    try:
                        date = self._get_iso_8601_week(
                            m.group('isoyear'),
                            m.group('isoweek'),
                            m.group('isoweekday')
                        )
                    except ParserError:
                        raise
                    except ValueError:
                        raise ParserError('Invalid date string: {}'.format(text))

                    year = date['year']
                    month = date['month']
                    day = date['day']
                else:
                    # We have a classic date representation
                    year = int(m.group('year'))

                    if not m.group('monthday'):
                        # No month and day
                        month = 1
                        day = 1
                    else:
                        if m.group('month') and m.group('day'):
                            # Month and day
                            if not m.group('daysep') and len(m.group('day')) == 1:
                                # Ordinal day
                                dt = datetime.strptime(
                                    '{}-{}'.format(year, m.group('month') + m.group('day')),
                                    '%Y-%j'
                                )
                                month = dt.month
                                day = dt.day
                            elif self._options['day_first']:
                                month = int(m.group('day'))
                                day = int(m.group('month'))
                            else:
                                month = int(m.group('month'))
                                day = int(m.group('day'))
                        else:
                            # Only month
                            if not m.group('monthsep'):
                                # The date looks like 201207
                                # which is invalid for a date
                                # But it might be a time in the form hhmmss
                                ambiguous_date = True

                            month = int(m.group('month'))
                            day = 1

                parsed.update({
                    'year': year,
                    'month': month,
                    'day': day,
                })

            if not m.group('time'):
                # No time has been specified
                if ambiguous_date:
                    # We can "safely" assume that the ambiguous date
                    # was actually a time in the form hhmmss
                    hhmmss = '{}{:0>2}'.format(
                        str(parsed['year']),
                        str(parsed['month'])
                    )

                    return {
                        'hour': int(hhmmss[:2]),
                        'minute': int(hhmmss[2:4]),
                        'second': int(hhmmss[4:]),
                        'subsecond': 0
                    }

                return parsed

            if ambiguous_date:
                raise ParserError('Invalid date string: {}'.format(text))

            if has_date and not m.group('timesep'):
                raise ParserError('Invalid date string: {}'.format(text))

            parsed.update({
                'hour': 0,
                'minute': 0,
                'second': 0,
                'subsecond': 0,
                'offset': None
            })

            # Grabbing hh:mm:ss
            parsed['hour'] = int(m.group('hour'))

            if m.group('minute'):
                parsed['minute'] = int(m.group('minute'))

            if m.group('second'):
                parsed['second'] = int(m.group('second'))

            # Grabbing subseconds, if any
            if m.group('subsecondsection'):
                # Limiting to 6 chars
                subsecond = m.group('subsecond')[:6]

                parsed['subsecond'] = int('{:0<6}'.format(subsecond))

            # Grabbing timezone, if any
            tz = m.group('tz')
            if tz:
                if tz == 'Z':
                    offset = 0
                else:
                    negative = True if tz.startswith('-') else False
                    tz = tz[1:]
                    if ':' not in tz:
                        if len(tz) == 2:
                            tz = '{}00'.format(tz)

                        off_hour = tz[0:2]
                        off_minute = tz[2:4]
                    else:
                        off_hour, off_minute = tz.split(':')

                    offset = ((int(off_hour) * 60) + int(off_minute)) * 60

                    if negative:
                        offset = -1 * offset

                parsed['offset'] = offset

            return parsed

    def _get_iso_8601_week(self, year, week, weekday):
        if not weekday:
            weekday = 1
        else:
            weekday = int(weekday)

        year = int(year)
        week = int(week)

        if week > 53:
            raise ParserError('Invalid week for week date')

        if weekday > 7:
            raise ParserError('Invalid weekday for week date')

        # We can't rely on strptime directly here since
        # it does not support ISO week date
        ordinal = week * 7 + weekday - (week_day(year, 1, 4) + 3)

        if ordinal < 1:
            # Previous year
            ordinal += days_in_year(year - 1)
            year -= 1

        if ordinal > days_in_year(year):
            # Next year
            ordinal -= days_in_year(year)
            year += 1

        fmt = '%Y-%j'
        string = '{}-{}'.format(year, ordinal)

        dt = datetime.strptime(string, fmt)

        return {
            'year': dt.year,
            'month': dt.month,
            'day': dt.day,
        }

    def parse(self, text):
        """
        Parses a string with the given options.

        :param text: The string to parse.
        :type text: str

        :rtype: dict
        """
        return self.normalize(self._parse(text))

    def normalize(self, parsed):
        """
        Normalizes the parsed element.

        :param parsed: The parsed elements.
        :type parsed: dict

        :rtype: dict
        """
        if self.is_exact():
            return parsed

        if any(('year' not in parsed, 'month' not in parsed, 'day' not in parsed)):
            now = self.now()
            default = {
                'year': now.year,
                'month': now.month,
                'day': now.day,
                'hour': 0,
                'minute': 0,
                'second': 0,
                'subsecond': 0,
                'offset': None
            }
        else:
            default = {
                'hour': 0,
                'minute': 0,
                'second': 0,
                'subsecond': 0,
                'offset': None
            }

        default.update(parsed)

        return default

    def _parse(self, text):
        # Trying to parse ISO8601 with C extension
        parsed = self._parse_iso8601(text)
        if parsed:
            return parsed

        parsed = self.parse_common(text)
        if parsed:
            return parsed

        # We couldn't parse the string
        # so we fallback on the dateutil parser
        try:
            dt = parser.parse(
                text,
                dayfirst=self._options['day_first'],
                yearfirst=self._options['year_first']
            )
        except ValueError:
            raise ParserError('Invalid date string: {}'.format(text))

        return {
            'year': dt.year,
            'month': dt.month,
            'day': dt.day,
            'hour': dt.hour,
            'minute': dt.minute,
            'second': dt.second,
            'subsecond': dt.microsecond,
            'offset': dt.utcoffset().total_seconds() if dt.tzinfo else None,
        }

    def _parse_iso8601(self, text):
        if not parse_iso8601:
            return

        try:
            dt = parse_iso8601(text, self._options['day_first'])
        except ValueError:
            return

        if isinstance(dt, time):
            return {
                'hour': dt.hour,
                'minute': dt.minute,
                'second': dt.second,
                'subsecond': dt.microsecond
            }
        elif isinstance(dt, date) and not isinstance(dt, datetime):
            return {
                'year': dt.year,
                'month': dt.month,
                'day': dt.day
            }

        parsed = {
            'year': dt.year,
            'month': dt.month,
            'day': dt.day,
            'hour': dt.hour,
            'minute': dt.minute,
            'second': dt.second,
            'subsecond': dt.microsecond,
            'offset': dt.tzinfo.offset if dt.tzinfo else None
        }

        return parsed



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta


class Transition(object):
    """
    Represents a Timezone transition time.

    It stores
        - the unix timestamp of the transition (in UTC),
        - the index of the TimezoneInfo (UTC offset,
                                         DST or not,
                                         timezone abbreviation),
        - the local time before the transition
        - the local time after the transition
        - the index of the previous TimezoneInfo
    """

    _epoch = datetime.utcfromtimestamp(0)

    def __init__(self, unix_time,
                 tzinfo_index, pre_time, time,
                 pre_tzinfo_index):
        """
        Constructor.

        :param unix_time: unix timestamp of the transition.
        :type unix_time: int

        :param tzinfo_index: The index of the timezone info
        :type tzinfo_index: int

        :param pre_time: The local time before the transition
        :type pre_time: datetime

        :param time: The local time after the transition
        :type time: datetime

        :param pre_tzinfo_index: The previous TimezoneInfo index
        :type pre_tzinfo_index: int
        """
        self._unix_time = unix_time
        self._tzinfo_index = tzinfo_index
        self._pre_time = pre_time
        self._time = time
        # We can't directly make datetime.utcfromtimestamp(unix_time)
        # since it will fail on Windows for negative timestamps.
        self._utc_time = self._epoch + timedelta(seconds=unix_time)
        self._pre_tzinfo_index = pre_tzinfo_index

    @property
    def unix_time(self):
        return self._unix_time

    @property
    def tzinfo_index(self):
        return self._tzinfo_index

    @property
    def pre_tzinfo_index(self):
        return self._pre_tzinfo_index

    @property
    def pre_time(self):
        return self._pre_time

    @property
    def time(self):
        return self._time

    @property
    def utc_time(self):
        return self._utc_time

    def __repr__(self):
        return '<Transition [{} UTC, {} -> {}]>'.format(
            self._utc_time,
            self._pre_time,
            self.time
        )



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'duminică',
        1: 'luni',
        2: 'marți',
        3: 'miercuri',
        4: 'joi',
        5: 'vineri',
        6: 'sâmbătă'
    },
    'days_abbrev': {
        0: 'Dum',
        1: 'Lun',
        2: 'Mar',
        3: 'Mie',
        4: 'Joi',
        5: 'Vin',
        6: 'Sâm'
    },

    # Months
    'months': {
        1: 'ianuarie',
        2: 'februarie',
        3: 'martie',
        4: 'aprilie',
        5: 'mai',
        6: 'iunie',
        7: 'iulie',
        8: 'august',
        9: 'septembrie',
        10: 'octombrie',
        11: 'noiembrie',
        12: 'decembrie'
    },
    'months_abbrev': {
        1: 'ian.',
        2: 'febr.',
        3: 'mart.',
        4: 'apr.',
        5: 'mai',
        6: 'iun.',
        7: 'iul.',
        8: 'aug.',
        9: 'sept.',
        10: 'oct.',
        11: 'nov.',
        12: 'dec.'
    },

    # Units of time
    'year': ['un an', '{count} ani', '{count} ani'],
    'month': ['o lună', '{count} luni', '{count} luni'],
    'week': ['o săptămână', '{count} săptămâni', '{count} săptămâni'],
    'day': ['o zi', '{count} zile', '{count} zile'],
    'hour': ['o oră', '{count} ore', '{count} ore'],
    'minute': ['un minut', '{count} minute', '{count} minute'],
    'second': ['o secundă', '{count} secunde', '{count} secunde'],

    # Relative time
    'ago': 'acum {time}',
    'from_now': '{time} de acum',
    'after': 'peste {time}',
    'before': 'acum {time}',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LT': 'H:mm',
        'LLLL': 'dddd, D MMMM YYYY H:mm',
        'LLL': 'D MMMM YYYY H:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'E Diel',
        1: 'E Hënë',
        2: 'E Martë',
        3: 'E Mërkurë',
        4: 'E Enjte',
        5: 'E Premte',
        6: 'E Shtunë'
    },
    'days_abbrev': {
        0: 'Die',
        1: 'Hën',
        2: 'Mar',
        3: 'Mër',
        4: 'Enj',
        5: 'Pre',
        6: 'Sht'
    },

    # Months
    'months': {
        1: 'Janar',
        2: 'Shkurt',
        3: 'Mars',
        4: 'Prill',
        5: 'Maj',
        6: 'Qershor',
        7: 'Korrik',
        8: 'Gusht',
        9: 'Shtator',
        10: 'Tetor',
        11: 'Nëntor',
        12: 'Dhjetor'
    },
    'months_abbrev': {
        1: 'Jan',
        2: 'Shk',
        3: 'Mar',
        4: 'Pri',
        5: 'Maj',
        6: 'Qer',
        7: 'Kor',
        8: 'Gus',
        9: 'Sht',
        10: 'Tet',
        11: 'Nën',
        12: 'Dhj'
    },

    # Units of time
    'year': ['{count} vit', '{count} vjet'],
    'month': ['{count} muaj', '{count} muaj'],
    'week': ['{count} javë', '{count} javë'],
    'day': ['{count} ditë', '{count} ditë'],
    'hour': ['{count} orë', '{count} orë'],
    'minute': ['{count} minutë', '{count} minuta'],
    'second': ['{count} sekondë', '{count} sekonda'],

    # Relative time
    'ago': '{time} më parë',
    'from_now': '{time} nga tani',
    'after': '{time} pas',
    'before': '{time} para',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'nedelja',
        1: 'ponedeljek',
        2: 'torek',
        3: 'sreda',
        4: 'četrtek',
        5: 'petek',
        6: 'sobota'
    },
    'days_abbrev': {
        0: 'ned.',
        1: 'pon.',
        2: 'tor.',
        3: 'sre.',
        4: 'čet.',
        5: 'pet.',
        6: 'sob.'
    },

    # Months
    'months': {
        1: 'januar',
        2: 'februar',
        3: 'marec',
        4: 'april',
        5: 'maj',
        6: 'junij',
        7: 'julij',
        8: 'avgust',
        9: 'september',
        10: 'oktober',
        11: 'november',
        12: 'december'
    },
    'months_abbrev': {
        1: 'jan.',
        2: 'feb.',
        3: 'mar.',
        4: 'apr.',
        5: 'maj.',
        6: 'jun.',
        7: 'jul.',
        8: 'avg.',
        9: 'sep.',
        10: 'okt.',
        11: 'nov.',
        12: 'dec.'
    },

    # Units of time
    'year': ['{count} leto', '{count} leti', '{count} leta', '{count} let'],
    'month': ['{count} mesec', '{count} meseca', '{count} mesece', '{count} mesecev'],
    'week': ['{count} teden', '{count} tedna', '{count} tedne', '{count} tednov'],
    'day': ['{count} dan', '{count} dni', '{count} dni', '{count} dni'],
    'hour': ['{count} uro', '{count} uri', '{count} ure', '{count} ur'],
    'minute': ['{count} minuto', '{count} minuti', '{count} minute', '{count} minut'],
    'second': ['{count} sekundo', '{count} sekundi', '{count} sekunde', '{count} sekund'],

    # Relative time
    'year_ago': ['{count} letom', '{count} leti', '{count} leti', '{count} leti'],
    'month_ago': ['{count} mesecem', '{count} meseci', '{count} meseci', '{count} meseci'],
    'week_ago': ['{count} tednom', '{count} tednoma', '{count} tedni', '{count} tedni'],
    'day_ago': ['{count} dnem', '{count} dnevoma', '{count} dnevi', '{count} dnevi'],
    'hour_ago': ['{count} uro', '{count} urama', '{count} urami', '{count} urami'],
    'minute_ago': ['{count} minuto', '{count} minutama', '{count} minutami', '{count} minutami'],
    'second_ago': ['{count} sekundo', '{count} sekundama', '{count} sekundami', '{count} sekundami'],
    'ago': 'pred {time}',
    'from_now': 'čez {time}',
    'after': 'čez {time}',
    'before': 'pred {time}',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LT': 'H:mm',
        'LLLL': 'dddd, D. MMMM YYYY H:mm',
        'LLL': 'D. MMMM YYYY H:mm',
        'LL': 'D. MMMM YYYY',
        'L': 'DD. MM. YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Sonntag',
        1: 'Montag',
        2: 'Dienstag',
        3: 'Mittwoch',
        4: 'Donnerstag',
        5: 'Freitag',
        6: 'Samstag'
    },
    'days_abbrev': {
        0: 'So',
        1: 'Mo',
        2: 'Di',
        3: 'Mi',
        4: 'Do',
        5: 'Fr',
        6: 'Sa'
    },

    # Months
    'months': {
        1: 'Januar',
        2: 'Februar',
        3: 'März',
        4: 'April',
        5: 'Mai',
        6: 'Juni',
        7: 'Juli',
        8: 'August',
        9: 'September',
        10: 'Oktober',
        11: 'November',
        12: 'Dezember',
    },
    'months_abbrev': {
        1: 'Jan',
        2: 'Feb',
        3: 'Mär',
        4: 'Apr',
        5: 'Mai',
        6: 'Jun',
        7: 'Jul',
        8: 'Aug',
        9: 'Sep',
        10: 'Okt',
        11: 'Nov',
        12: 'Dez',
    },

    # Units of time
    'year': ['{count} Jahr', '{count} Jahre'],
    'month': ['{count} Monat', '{count} Monate'],
    'week': ['{count} Woche', '{count} Wochen'],
    'day': ['{count} Tag', '{count} Tage'],
    'hour': ['{count} Stunde', '{count} Stunden'],
    'minute': ['{count} Minute', '{count} Minuten'],
    'second': ['{count} Sekunde', '{count} Sekunden'],

    # Relative time
    'ago': 'vor {time}',
    'from_now': 'in {time}',
    'after': '{time} später',
    'before': '{time} zuvor',

    'year_from_now': ['{count} Jahr', '{count} Jahren'],
    'month_from_now': ['{count} Monat', '{count} Monaten'],
    'week_from_now': ['{count} Woche', '{count} Wochen'],
    'day_from_now': ['{count} Tag', '{count} Tagen'],
    'year_ago': ['{count} Jahr', '{count} Jahren'],
    'month_ago': ['{count} Monat', '{count} Monaten'],
    'week_ago': ['{count} Woche', '{count} Wochen'],
    'day_ago': ['{count} Tag', '{count} Tagen'],

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D. MMMM YYYY HH:mm',
        'LLL': 'D. MMMM YYYY HH:mm',
        'LL': 'D. MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

from datetime import timedelta

from .mixins.interval import (
    WordableIntervalMixin
)

from .constants import (
    SECONDS_PER_DAY, SECONDS_PER_HOUR,
    SECONDS_PER_MINUTE
)


def _divide_and_round(a, b):
    """divide a by b and round result to the nearest integer

    When the ratio is exactly half-way between two integers,
    the even integer is returned.
    """
    # Based on the reference implementation for divmod_near
    # in Objects/longobject.c.
    q, r = divmod(a, b)
    # round up if either r / b > 0.5, or r / b == 0.5 and q is odd.
    # The expression r / b > 0.5 is equivalent to 2 * r > b if b is
    # positive, 2 * r < b if b negative.
    r *= 2
    greater_than_half = r > b if b > 0 else r < b
    if greater_than_half or r == b and q % 2 == 1:
        q += 1

    return q


class BaseInterval(timedelta):
    """
    Base class for all inherited interval classes.
    """

    _y = None
    _m = None
    _w = None
    _d = None
    _h = None
    _i = None
    _s = None
    _invert = None

    def __new__(cls, days=0, seconds=0, microseconds=0,
                milliseconds=0, minutes=0, hours=0, weeks=0):
        self = timedelta.__new__(
            cls, days, seconds, microseconds,
            milliseconds, minutes, hours, weeks
        )

        # Intuitive normalization
        total = self.total_seconds()

        m = 1
        if total < 0:
            m = -1

        self._microseconds = round(total % m * 1e6)
        self._seconds = abs(int(total)) % SECONDS_PER_DAY * m
        self._days = abs(int(total)) // SECONDS_PER_DAY * m

        return self

    def total_minutes(self):
        return self.total_seconds() / SECONDS_PER_MINUTE

    def total_hours(self):
        return self.total_seconds() / SECONDS_PER_HOUR

    def total_days(self):
        return self.total_seconds() / SECONDS_PER_DAY

    def total_weeks(self):
        return self.total_days() / 7

    @property
    def weeks(self):
        return abs(self.days) // 7 * self._sign(self._days)

    @property
    def days(self):
        return self._days

    @property
    def remaining_days(self):
        return abs(self._days) % 7 * self._sign(self._days)

    @property
    def hours(self):
        if self._h is None:
            seconds = self._seconds
            self._h = 0
            if abs(seconds) >= 3600:
                self._h = (abs(seconds) // 3600 % 24) * self._sign(seconds)

        return self._h

    @property
    def minutes(self):
        if self._i is None:
            seconds = self._seconds
            self._i = 0
            if abs(seconds) >= 60:
                self._i = (abs(seconds) // 60 % 60) * self._sign(seconds)

        return self._i

    @property
    def seconds(self):
        return self._seconds

    @property
    def remaining_seconds(self):
        if self._s is None:
            self._s = self._seconds
            self._s = abs(self._s) % 60 * self._sign(self._s)

        return self._s

    @property
    def microseconds(self):
        return self._microseconds

    @property
    def invert(self):
        if self._invert is None:
            self._invert = self.total_seconds() < 0

        return self._invert

    def in_weeks(self):
        return int(self.total_weeks())

    def in_days(self):
        return int(self.total_days())

    def in_hours(self):
        return int(self.total_hours())

    def in_minutes(self):
        return int(self.total_minutes())

    def in_seconds(self):
        return int(self.total_seconds())

    def _sign(self, value):
        if value < 0:
            return -1

        return 1

    def as_timedelta(self):
        """
        Return the interval as a native timedelta.

        :rtype: timedelta
        """
        return timedelta(seconds=self.total_seconds())


class Interval(WordableIntervalMixin, BaseInterval):
    """
    Replacement for the standard timedelta class.

    Provides several improvements over the base class.
    """

    @classmethod
    def instance(cls, delta):
        """
        Creates a Interval from a timedelta

        :type delta: timedelta

        :rtype: Interval
        """
        return cls(days=delta.days, seconds=delta.seconds, microseconds=delta.microseconds)

    def __add__(self, other):
        if isinstance(other, timedelta):
            return self.__class__(seconds=self.total_seconds() + other.total_seconds())

        return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, timedelta):
            return self.__class__(seconds=self.total_seconds() - other.total_seconds())

        return NotImplemented

    def __neg__(self):
        return self.__class__(seconds=-self.total_seconds())

    def _to_microseconds(self):
        return ((self._days * (24*3600) + self._seconds) * 1000000 +
                self._microseconds)

    def __mul__(self, other):
        if isinstance(other, int):
            return self.__class__(seconds=self.total_seconds() * other)

        if isinstance(other, float):
            usec = self._to_microseconds()
            a, b = other.as_integer_ratio()

            return self.__class__(0, 0, _divide_and_round(usec * a, b))

        return NotImplemented

    __rmul__ = __mul__

    def __floordiv__(self, other):
        if not isinstance(other, (int, timedelta)):
            return NotImplemented

        usec = self._to_microseconds()
        if isinstance(other, timedelta):
            return usec // other._to_microseconds()

        if isinstance(other, int):
            return self.__class__(0, 0, usec // other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float, timedelta)):
            return NotImplemented

        usec = self._to_microseconds()
        if isinstance(other, timedelta):
            return usec / other._to_microseconds()

        if isinstance(other, int):
            return self.__class__(0, 0, _divide_and_round(usec, other))

        if isinstance(other, float):
            a, b = other.as_integer_ratio()

            return self.__class__(0, 0, _divide_and_round(b * usec, a))

    __div__ = __floordiv__

    def __mod__(self, other):
        if isinstance(other, timedelta):
            r = self._to_microseconds() % other._to_microseconds()

            return self.__class__(0, 0, r)

        return NotImplemented

    def __divmod__(self, other):
        if isinstance(other, timedelta):
            q, r = divmod(self._to_microseconds(),
                          other._to_microseconds())

            return q, self.__class__(0, 0, r)

        return NotImplemented

Interval.min = Interval(-999999999)
Interval.max = Interval(days=999999999, hours=23,
                        minutes=59, seconds=59,
                        microseconds=999999)
Interval.resolution = Interval(microseconds=1)


class AbsoluteInterval(Interval):
    """
    Interval that expresses a time difference in absolute values.
    """

    def __new__(cls, days=0, seconds=0, microseconds=0,
                milliseconds=0, minutes=0, hours=0, weeks=0):
        self = timedelta.__new__(
            cls, days, seconds, microseconds,
            milliseconds, minutes, hours, weeks
        )

        # We need to compute the total_seconds() value
        # on a native timedelta object
        delta = timedelta(
            days, seconds, microseconds,
            milliseconds, minutes, hours, weeks
        )

        # Intuitive normalization
        self._total = delta.total_seconds()
        total = abs(self._total)

        self._microseconds = round(total % 1 * 1e6)
        self._seconds = int(total) % SECONDS_PER_DAY
        self._days = int(total) // SECONDS_PER_DAY

        return self

    def total_seconds(self):
        return abs(self._total)

    @property
    def invert(self):
        if self._invert is None:
            self._invert = self._total < 0

        return self._invert



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Κυριακή',
        1: 'Δευτέρα',
        2: 'Τρίτη',
        3: 'Τετάρτη',
        4: 'Πέμπτη',
        5: 'Παρασκευή',
        6: 'Σάββατο'
    },
    'days_abbrev': {
        0: 'Κυρ',
        1: 'Δευ',
        2: 'Τρι',
        3: 'Τετ',
        4: 'Πεμ',
        5: 'Παρ',
        6: 'Σαβ'
    },

    # Months
    'months': {
        1: 'Ιανουαρίου',
        2: 'Φεβρουαρίου',
        3: 'Μαρτίου',
        4: 'Απριλίου',
        5: 'Μαΐου',
        6: 'Ιουνίου',
        7: 'Ιουλίου',
        8: 'Αυγούστου',
        9: 'Σεπτεμβρίου',
        10: 'Οκτωβρίου',
        11: 'Νοεμβρίου',
        12: 'Δεκεμβρίου',
    },
    'months_abbrev': {
        1: 'Ιαν',
        2: 'Φεβ',
        3: 'Μαρ',
        4: 'Απρ',
        5: 'Μαϊ',
        6: 'Ιον',
        7: 'Ιολ',
        8: 'Αυγ',
        9: 'Σεπ',
        10: 'Οκτ',
        11: 'Νοε',
        12: 'Δεκ',
    },

    # Units of time
    'year': ['{count} χρόνος', '{count} χρόνια'],
    'month': ['{count} μήνας', '{count} μήνες'],
    'week': ['{count} εβδομάδα', '{count} εβδομάδες'],
    'day': ['{count} μέρα', '{count} μέρες'],
    'hour': ['{count} ώρα', '{count} ώρες'],
    'minute': ['{count} λεπτό', '{count} λεπτά'],
    'second': ['{count} δευτερόλεπτο', '{count} δευτερόλεπτα'],

    # Relative time
    'ago': 'πρίν απο {time}',
    'from_now': 'σε {time} απο τώρα',
    'after': '{time} μετά',
    'before': '{time} πρίν',

    # Date formats
    'date_formats': {
        'LTS': 'h:mm:ss A',
        'LT': 'h:mm A',
        'LLLL': 'dddd, D MMMM YYYY h:mm A',
        'LLL': 'D MMMM YYYY h:mm A',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
# Days
    'days': {
        0: 'Diumenge',
        1: 'Dilluns',
        2: 'Dimarts',
        3: 'Dimecres',
        4: 'Dijous',
        5: 'Divendres',
        6: 'Dissabte'
    },
    'days_abbrev': {
        0: 'Diumenge',
        1: 'Dilluns',
        2: 'Dimarts',
        3: 'Dimecres',
        4: 'Dijous',
        5: 'Divendres',
        6: 'Dissabte'
    },

    # Months
    'months': {
        1: 'Gener',
        2: 'Febrer',
        3: 'Març',
        4: 'Abril',
        5: 'Maig',
        6: 'Juny',
        7: 'Juliol',
        8: 'Agost',
        9: 'Setembre',
        10: 'Octubre',
        11: 'Novembre',
        12: 'Desembre',
    },
    'months_abbrev': {
        1: 'Gener',
        2: 'Febrer',
        3: 'Març',
        4: 'Abril',
        5: 'Maig',
        6: 'Juny',
        7: 'Juliol',
        8: 'Agost',
        9: 'Setembre',
        10: 'Octubre',
        11: 'Novembre',
        12: 'Desembre',
    },

    # Units of time
    'year': ['{count} any', '{count} anys'],
    'month': ['{count} mes', '{count} mesos'],
    'week': ['{count} setmana', '{count} setmanes'],
    'day': ['{count} dia', '{count} díes'],
    'hour': ['{count} hora', '{count} hores'],
    'minute': ['{count} minut', '{count} minuts'],
    'second': ['{count} segon', '{count} segons'],

    # Relative time
    'ago': 'Fa {time}',
    'from_now': 'Dins de {time}',
    'after': '{time} després',
    'before': '{time} abans',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LT': 'H:mm',
        'LLLL': 'dddd D MMMM YYYY H:mm',
        'LLL': 'D MMMM YYYY H:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'sekmadienis',
        1: 'pirmadienis',
        2: 'antradienis',
        3: 'trečiadienis',
        4: 'ketvirtadienis',
        5: 'penktadienis',
        6: 'šeštadienis'
    },
    'days_abbrev': {
        0: 'Sek',
        1: 'Pir',
        2: 'Ant',
        3: 'Tre',
        4: 'Ket',
        5: 'Pen',
        6: 'Šeš'
    },

    # Months
    'months': {
        1: 'sausis',
        2: 'vasaris',
        3: 'kovas',
        4: 'balandis',
        5: 'gegužė',
        6: 'birželis',
        7: 'liepa',
        8: 'rugpjūtis',
        9: 'rugsėjis',
        10: 'spalis',
        11: 'lapkritis',
        12: 'gruodis'
    },
    'months_abbrev': {
        1: 'sau',
        2: 'vas',
        3: 'kov',
        4: 'bal',
        5: 'geg',
        6: 'bir',
        7: 'lie',
        8: 'rgp',
        9: 'rgs',
        10: 'spa',
        11: 'lap',
        12: 'grd'
    },

    # Units of time
    'year': ['{count} metus', '{count} metus', '{count} metų'],
    'month': ['{count} mėnesį', '{count} mėnesius', '{count} mėnesių'],
    'week': ['{count} savaitę', '{count} savaites', '{count} savaičių'],
    'day': ['{count} dieną', '{count} dienas', '{count} dienų'],
    'hour': ['{count} valandą', '{count} valandas', '{count} valandų'],
    'minute': ['{count} minutę', '{count} minutes', '{count} minučių'],
    'second': ['{count} sekundę', '{count} sekundes', '{count} sekundžių'],

    # Relative time
    'second_from_now': ['{count} sekundės', '{count} sekundžių', '{count} sekundžių'],
    'minute_from_now': ['{count} minutės', '{count} minučių', '{count} minučių'],
    'hour_from_now': ['{count} valandos', '{count} valandų', '{count} valandų'],
    'day_from_now': ['{count} dienos', '{count} dienų', '{count} dienų'],
    'week_from_now': ['{count} savaitės', '{count} savaičių', '{count} savaičių'],
    'month_from_now': ['{count} mėnesio', '{count} mėnesių', '{count} mėnesių'],
    'year_from_now': 'metų',
    'ago': 'prieš {time}',
    'from_now': 'už {time}',
    'after': 'po {time}',
    'before': '{time} nuo dabar',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'YYYY [m.] MMMM D [d.], dddd, HH:mm [val.]',
        'LLL': 'YYYY [m.] MMMM D [d.], HH:mm [val.]',
        'LL': 'YYYY [m.] MMMM D [d.]',
        'L': 'YYYY-MM-DD',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'یکشنبه',
        1: 'دو شنبه',
        2: 'سه شنبه',
        3: 'چهارشنبه',
        4: 'پنجشنبه',
        5: 'جمعه',
        6: 'شنبه'
    },
    'days_abbrev': {
        0: 'یکشنبه',
        1: 'دوشنبه',
        2: 'سهشنبه',
        3: 'چهارشنبه',
        4: 'پنجشنبه',
        5: 'جمعه',
        6: 'شنبه'
    },

    # Months
    'months': {
        1: 'ژانویه',
        2: 'فوریه',
        3: 'مارس',
        4: 'آوریل',
        5: 'مه',
        6: 'ژوئن',
        7: 'ژوئیه',
        8: 'اوت',
        9: 'سپتامبر',
        10: 'اکتبر',
        11: 'نوامبر',
        12: 'دسامبر'
    },
    'months_abbrev': {
        1: 'ژانویه',
        2: 'فوریه',
        3: 'مارس',
        4: 'آوریل',
        5: 'مه',
        6: 'ژوئن',
        7: 'ژوئیه',
        8: 'اوت',
        9: 'سپتامبر',
        10: 'اکتبر',
        11: 'نوامبر',
        12: 'دسامبر',
    },

    # Units of time
    'year': '{count} سال',
    'month': '{count} ماه',
    'week': '{count} هفته',
    'day': '{count} روز',
    'hour': '{count} ساعت',
    'minute': '{count} دقیقه',
    'second': '{count} ثانیه',

    # Relative time
    'ago': '{time} پیش',
    'from_now': '{time} بعد',
    'after': '{time} پس از',
    'before': '{time} پیش از',

    # Meridians
    'meridian': lambda time: 'قبل از ظهر' if 0 <= time[0] < 12 else 'بعد از ظهر',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

import pendulum

from math import copysign
from datetime import datetime, timedelta

try:
    from ._extensions._helpers import local_time, parse_iso8601 as _parse_iso8601

    def parse_iso8601(text, day_first=False):
        return _parse_iso8601(text, day_first)

except ImportError:
    from ._extensions.helpers import local_time

    parse_iso8601 = None

from .constants import (
    DAYS_PER_MONTHS, DAY_OF_WEEK_TABLE, DAYS_PER_L_YEAR, DAYS_PER_N_YEAR
)


def is_leap(year):
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def add_duration(dt, years=0, months=0, weeks=0, days=0,
                 hours=0, minutes=0, seconds=0, microseconds=0):
    """
    Adds a duration to a datetime instance.

    :param dt: The datetime instance
    :type dt: datetime.datetime

    :param years: The number of years
    :type years: int

    :param months: The number of months
    :type months: int

    :param weeks: The number of weeks
    :type weeks: int

    :param days: The number of days
    :type days: int

    :param hours: The number of hours
    :type hours: int

    :param minutes: The number of minutes
    :type minutes: int

    :param seconds: The number of seconds
    :type seconds: int

    :param microseconds: The number of microseconds
    :type microseconds: int

    :rtype: datetime.datetime
    """
    days += weeks * 7

    # Normalizing
    if abs(microseconds) > 999999:
        s = _sign(microseconds)
        div, mod = divmod(microseconds * s, 1000000)
        microseconds = mod * s
        seconds += div * s

    if abs(seconds) > 59:
        s = _sign(seconds)
        div, mod = divmod(seconds * s, 60)
        seconds = mod * s
        minutes += div * s

    if abs(minutes) > 59:
        s = _sign(minutes)
        div, mod = divmod(minutes * s, 60)
        minutes = mod * s
        hours += div * s

    if abs(hours) > 23:
        s = _sign(hours)
        div, mod = divmod(hours * s, 24)
        hours = mod * s
        days += div * s

    if abs(months) > 11:
        s = _sign(months)
        div, mod = divmod(months * s, 12)
        months = mod * s
        years += div * s

    year = dt.year + years
    month = dt.month

    if months:
        month += months
        if month > 12:
            year += 1
            month -= 12
        elif month < 1:
            year -= 1
            month += 12

    day = min(DAYS_PER_MONTHS[int(is_leap(year))][month], dt.day)

    dt = dt.replace(year=year, month=month, day=day)

    return dt + timedelta(
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        microseconds=microseconds
    )


def precise_diff(d1, d2):
    """
    Calculate a precise difference between two datetimes.

    :param d1: The first datetime
    :type d1: pendulum.Pendulum or pendulum.Date

    :param d2: The second datetime
    :type d2: pendulum.Pendulum or pendulum.Date

    :rtype: dict
    """
    diff = {
        'years': 0,
        'months': 0,
        'days': 0,
        'hours': 0,
        'minutes': 0,
        'seconds': 0,
        'microseconds': 0,
        'total': {
            'days': 0
        }
    }
    sign = 1

    if d1 == d2:
        return diff

    tzinfo1 = d1.tzinfo if isinstance(d1, datetime) else None
    tzinfo2 = d2.tzinfo if isinstance(d2, datetime) else None

    if (tzinfo1 is None and tzinfo2 is not None
        or tzinfo2 is None and tzinfo1 is not None):
        raise ValueError(
            'Comparison between naive and aware datetimes is not supported'
        )

    if d1 > d2:
        d1, d2 = d2, d1
        sign = -1

    d_diff = 0
    hour_diff = 0
    min_diff = 0
    sec_diff = 0
    mic_diff = 0
    total_days = (
        _day_number(d2.year, d2.month, d2.day)
        - _day_number(d1.year, d1.month, d1.day)
    )

    in_same_tz = False
    tz1 = None
    tz2 = None

    # Trying to figure out the timezone names
    # If we can't find them, we assume different timezones
    if tzinfo1 and tzinfo2:
        if hasattr(tzinfo1, 'name'):
            # Pendulum timezone
            tz1 = tzinfo1.name
        elif hasattr(tzinfo1, 'zone'):
            # pytz timezone
            tz1 = tzinfo1.zone

        if hasattr(tzinfo2, 'name'):
            tz2 = tzinfo2.name
        elif hasattr(tzinfo2, 'zone'):
            tz2 = tzinfo2.zone

        in_same_tz = tz1 == tz2 and tz1 is not None

    if hasattr(d2, 'hour'):
        # If we are not in the same timezone
        # we need to adjust
        #
        # We also need to adjust if we do not
        # have variable-length units
        if not in_same_tz or total_days == 0:
            offset1 = d1.utcoffset()
            offset2 = d2.utcoffset()

            if offset1:
                d1 = d1 - offset1

            if offset2:
                d2 = d2 - offset2

        hour_diff = d2.hour - d1.hour
        min_diff = d2.minute - d1.minute
        sec_diff = d2.second - d1.second
        mic_diff = d2.microsecond - d1.microsecond

        if mic_diff < 0:
            mic_diff += 1000000
            sec_diff -= 1

        if sec_diff < 0:
            sec_diff += 60
            min_diff -= 1

        if min_diff < 0:
            min_diff += 60
            hour_diff -= 1

        if hour_diff < 0:
            hour_diff += 24
            d_diff -= 1

    if d1 > d2:
        d1, d2 = d2, d1
        sign = -1

    y_diff = d2.year - d1.year
    m_diff = d2.month - d1.month
    d_diff += d2.day - d1.day

    if d_diff < 0:
        year = d2.year
        month = d2.month

        if month == 1:
            month = 12
            year -= 1
        else:
            month -= 1

        leap = int(is_leap(year))

        days_in_last_month = DAYS_PER_MONTHS[leap][month]
        days_in_month = DAYS_PER_MONTHS[int(is_leap(d2.year))][d2.month]

        if d_diff < days_in_month - days_in_last_month:
            # We don't have a full month, we calculate days
            if days_in_last_month < d1.day:
                d_diff += d1.day
            else:
                d_diff += days_in_last_month
        elif d_diff == days_in_month - days_in_last_month:
            # We have exactly a full month
            # We remove the days difference
            # and add one to the months difference
            d_diff = 0
            m_diff += 1
        else:
            # We have a full month
            d_diff += days_in_last_month

        m_diff -= 1

    if m_diff < 0:
        m_diff += 12
        y_diff -= 1

    diff['microseconds'] = sign * mic_diff
    diff['seconds'] = sign * sec_diff
    diff['minutes'] = sign * min_diff
    diff['hours'] = sign * hour_diff
    diff['days'] = sign * d_diff
    diff['months'] = sign * m_diff
    diff['years'] = sign * y_diff
    diff['total']['days'] = sign * total_days

    return diff


def week_day(year, month, day):
    if month < 3:
        year -= 1

    w = (year + year//4 - year//100 + year//400 + DAY_OF_WEEK_TABLE[month - 1] + day) % 7

    if not w:
        w = 7

    return w


def days_in_year(year):
    if is_leap(year):
        return DAYS_PER_L_YEAR

    return DAYS_PER_N_YEAR


def _sign(x):
    return int(copysign(1, x))


def _day_number(year, month, day):
    month = (month + 9) % 12
    year = year - month // 10

    return (
        365 * year
        + year // 4 - year // 100 + year // 400
        + (month * 306 + 5) // 10
        + (day - 1 )
    )
